package dictionarybuilder;

import java.util.*;

import java.io.*;

/**
 * 
 *
 * @author quasiyoke
 */
public class Index implements Comparable{

/**
 * Символ, использующийся для обозначения индекса, которому принадлежит
 * любое количество любых подиндексов.
 */
private final static char FREE_CHARACTER='+';

public static final int HAVE_NOT_FREE_SUBINDEX_MASK=0x80000000;

/**
 * Массив символов, которые не имеют значения при индексировании слов.
 * Сортируется в <code>static{}</code> блоке для дальнейшего поиска внутри
 * него.
 */
private final static char[] SPACE_CHARACTERS={0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,
    15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,' ','-','.',',','\'','?',
    '!',FREE_CHARACTER,
};

public static final int SUBINDEXES_OFFSET_MASK=0x7FFFFFFF;

private static Vector<Character> alphabet;

private int articlesCount;

private int articleOffset;

/**
 * Буква, представляемая индексом.
 */
private char character;

/**
 * Которую по счёту букву в ключе представляет этот индекс.
 */
private int depth;

/**
 * Есть ли у индекса подиндекс с любым числом любых букв -- свободный
 * подиндекс.
 */
private boolean haveFreeSubindex;

private Index parent;

/**
 * Смещение места в результирующем файле (внутри индекса-родителя), в
 * котором должно оказаться смещение на этот индекс.
 */
private int pointerToMeOffset;

private static Index root;

private Hashtable<Character,Index> subindex;


static{
    Arrays.sort(SPACE_CHARACTERS);
    alphabet=new Vector<Character>();
    root=new Index();
}

private Index(){
    articlesCount=0;
    articleOffset=pointerToMeOffset=Main.NULL_POINTER;
    subindex=new Hashtable<Character,Index>();
    depth=-1;
    character=0;
    parent=null;
}

private Index(String name,Index parent){
    articlesCount=0;
    articleOffset=pointerToMeOffset=Main.NULL_POINTER;
    subindex=new Hashtable<Character,Index>();
    character=name.charAt(depth=parent.depth+1);
    this.parent=parent;
}

private Index(char character,Index parent){
    articlesCount=0;
    articleOffset=pointerToMeOffset=Main.NULL_POINTER;
    subindex=new Hashtable<Character,Index>();
    this.character=character;
    this.depth=parent.depth+1;
    this.parent=parent;
}

/**
 * @param articleOffset
 *            Смещение этой статьи от первого байта первой статьи.
 */
public void addArticleOffset(int articleOffset){
    /*
    System.out.println(articleOffset);
     */
    this.articleOffset=articleOffset;
    incrementArticlesCount();
}

public static void addToAlphabet(char character){
    if(Arrays.binarySearch(SPACE_CHARACTERS,character)>=0)
	return;
    character=Character.toLowerCase(character);
    if(!alphabet.contains(character))
	alphabet.add(character);
}

public static void addToAlphabet(int index,char character){
    if(Arrays.binarySearch(SPACE_CHARACTERS,character)>=0)
	return;
    character=Character.toLowerCase(character);
    if(!alphabet.contains(character)){
	if(alphabet.size()<=index)
	    alphabet.setSize(index+1);
	alphabet.set(index,character);
    }
}

/**
 * Добавлет в конец данного строкового буфера имя индекса.
 */
private void appendName(StringBuffer stringBuffer){
    if(parent==null)
	return;
    parent.appendName(stringBuffer);
    stringBuffer.append(character);
}

public int compareTo(Object object){
    if(object instanceof Index){
	Index index=(Index)object;
	return (this.articlesCount>index.articlesCount||this.character==
		FREE_CHARACTER)?-1:(this.articlesCount<index.articlesCount||
		index.character==FREE_CHARACTER)?1:0;
    }
    return -1;
}

/**
 * @param name
 *            Имя индекса. Всегда проходит {@link prepareString(String)}.
 * @param createIfAny
 *            Допускается ли создание индекса в случае отсутствия нужного.
 * @return Индекс с данным именем.
 */
public static Index[] get(String name,boolean createIfAny){
    int j;
    name=name.trim().toLowerCase();
    int nameLength=name.length();
    StringBuffer prefix=new StringBuffer();
    StringBuffer body=new StringBuffer();
    StringBuffer suffix=new StringBuffer();
    StringBuffer stringBuffer;
    boolean haveBody=false;
    char character;
    boolean bodyClosed=false;
    for(j=0;j<nameLength;j++){
	character=name.charAt(j);
	if(Arrays.binarySearch(SPACE_CHARACTERS,character)>=0)
	    continue;
	if(haveBody){
	    if(character==')'){
		bodyClosed=true;
		continue;
	    }
	}else if(haveBody=(character=='('))
	    continue;
	stringBuffer=haveBody?(bodyClosed?suffix:body):prefix;
	if(!alphabet.contains(character)){
	    stringBuffer.append(FREE_CHARACTER);
	    System.err.println("Error, character \'"+character+"\' ("+
		    (int)character+", 0x"+Integer.toHexString(
		    character).toUpperCase()+", from \""+name+
		    "\", offset "+j+") not found in alphabet.");
	    break;
	}
	stringBuffer.append(character);
    }
    prefix.append(suffix);
    if(haveBody){
	Index[] index={root.getSubindex(prefix.toString(),createIfAny),root.
	    getSubindex(prefix.insert(prefix.length()-suffix.length(),body).
	    toString(),
	    createIfAny),
	};
	return index;
    }else{
	Index[] index={root.getSubindex(prefix.toString(),createIfAny),};
	return index;
    }
}

public static int getAlphabetSize(){
    return alphabet.size();
}

public static int getArticlesCount(){
    return root.articlesCount;
}

private Index getSubindex(String name,boolean createIfAny){
    int index=depth+1;
    int nameLength=name.length();
    if(nameLength==index)
	return this;
    else if(nameLength<index)
	return null;
    char character=name.charAt(index);
    Index aSubindex=subindex.get(character);
    if(aSubindex==null)
	if(createIfAny){
	    aSubindex=new Index(character,this);
	    subindex.put(character,aSubindex);
	    if(character==FREE_CHARACTER)
		haveFreeSubindex=true;
	}else
	    return null;
    return aSubindex.getSubindex(name,createIfAny);
}

private void incrementArticlesCount(){
    articlesCount++;
    if(parent!=null)
	parent.incrementArticlesCount(this);
}

private void incrementArticlesCount(Index subindex){
    if(subindex.depth==this.depth+1){
	if(!this.subindex.containsValue(subindex))
	    this.subindex.put(subindex.character,subindex);
	incrementArticlesCount();
    }
}

/**
 * @return Длину индекса и всех его подиндексов в байтах.
 */
private int length(){
    int subindexCount=subindex.size();
    int length=length(subindexCount);
    Enumeration eSubindex=subindex.elements();
    while(eSubindex.hasMoreElements())
	length+=((Index)eSubindex.nextElement()).length();
    return length;
}

/**
 * @param subindexCount
 *            Число подиндексов.
 * @return Длину индекса в байтах.
 */
public static int length(int subindexCount){
    return Main.INT_LENGTH*3+Main.BYTE_LENGTH*2*subindexCount;
}

/**
 * @param args 
 *            Аргументы командной строки.
 */
public static void main(String[] args){
}

public String toString(){
    StringBuffer stringBuffer=new StringBuffer();
    if(depth==-1)
	stringBuffer.append("root");
    else{
	stringBuffer.append('\"');
	appendName(stringBuffer);
	stringBuffer.append("\", depth ");
	stringBuffer.append(depth);
    }
    stringBuffer.append(" (");
    stringBuffer.append(this.articlesCount);
    stringBuffer.append(" art.):");
    Enumeration eSubindex=subindex.elements();
    while(eSubindex.hasMoreElements()){
	stringBuffer.append(" \'");
	stringBuffer.append(((Index)eSubindex.nextElement()).character);
	stringBuffer.append('\'');
    }
    return stringBuffer.toString();
}

/**
 * @param articlesOffset
 *            Смещение первого байта первой статьи от начала конечного
 *            файла.
 * @return Подиндексы этого индекса.
 */
private Index[] write(EasyAccessFile dictionaryAccess,
	int articlesOffset) throws IOException{
    if(pointerToMeOffset!=Main.NULL_POINTER){
	int offset=(int)dictionaryAccess.getFilePointer();
	dictionaryAccess.seek(pointerToMeOffset);
	dictionaryAccess.writeInt(offset|
		(parent.haveFreeSubindex?0:HAVE_NOT_FREE_SUBINDEX_MASK));
	dictionaryAccess.seek(offset);
    }
    dictionaryAccess.writeInt(this.articlesCount);
    dictionaryAccess.writeInt(this.articleOffset==Main.NULL_POINTER?Main.NULL_POINTER
	    :articlesOffset+this.articleOffset);
    Index[] arSubindex=subindex.values().toArray(new Index[0]);
    Arrays.sort(arSubindex);
    if(arSubindex.length!=0)
	arSubindex[0].pointerToMeOffset=(int)dictionaryAccess.getFilePointer();
    dictionaryAccess.writeInt(Main.NULL_POINTER);
    int index;
    char character;
    int j;
    for(j=0;j<arSubindex.length;j++){
	character=arSubindex[j].character;
	if(character==FREE_CHARACTER)
	    index=0xFF;
	else if((index=alphabet.indexOf(character))==-1){
	    System.err.println("Error, character #"+(int)character+" (0x"+Integer.toHexString(character).
		    toUpperCase()+") of index\n\t"+arSubindex[j].toString()+
		    "\nnot found in alphabet.");
	    continue;
	}
	dictionaryAccess.writeByte(index);
	dictionaryAccess.writeByte(arSubindex[j].subindex.size());
    }
    /*
    if(haveFreeSubindex)
    haveFreeSubindex=haveFreeSubindex;
     */
    return arSubindex;
}

/**
 * Записывает в поток алфавит и все индексы.
 */
public static void writeAll(EasyAccessFile dictionaryAccess,
	String dictionaryKeysLanguage,String dictionaryArticlesLanguage,
	String dictionarySource)
	throws IOException{
    int alphabetSize=alphabet.size();
    int articlesOffset=dictionaryAccess.writeUTF8String(
	    dictionaryKeysLanguage)+dictionaryAccess.writeUTF8String(
	    dictionaryArticlesLanguage)+dictionaryAccess.writeUTF8String(
	    dictionarySource)+Main.BYTE_LENGTH*2+Main.CHAR_LENGTH*alphabetSize+
	    root.length();
    dictionaryAccess.writeByte(alphabetSize);
    int j;
    int i;
    for(j=0;j<alphabetSize;j++)
	dictionaryAccess.writeChar(alphabet.get(j));
    dictionaryAccess.writeByte(root.subindex.size());
    Vector<Index> subindex=new Vector<Index>();
    Vector<Index> index=new Vector<Index>();
    Index[] arSubindex;
    index.add(root);
    Vector<Index> temp;
    int indexesCount;
    do{
	subindex.clear();
	indexesCount=index.size();
	for(j=0;j<indexesCount;j++){
	    arSubindex=index.get(j).write(dictionaryAccess,articlesOffset);
	    for(i=0;i<arSubindex.length;i++)
		subindex.add(arSubindex[i]);
	}
	temp=subindex;
	subindex=index;
	index=temp;
    }while(!index.isEmpty());
}

}
